-- name: Duels v1.2
-- description: Players may use the /duel command to duel each other.\n\nMod originally made for BizzareScape by EmilyEmmi.

LEVEL_ARENA = level_register("level_arena_entry", 0, "Battle Arena", "BA", 28000, 0x28, 0x28, 0x28)
ACT_DO_NOTHING = allocate_mario_action(ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE)
ACT_DEBUG_FREE_MOVE_CUSTOM = allocate_mario_action(ACT_GROUP_CUTSCENE | ACT_FLAG_INTANGIBLE)

local opponentDCTimer = 60
local didStart = false
local betweenTimer = 0
local TIME_STOP_MAX = 3
local time_stop = TIME_STOP_MAX
local hostIndex = -1
local prevLevel, prevAct, prevArea = 0, 0, 0
local requestData = {}
for i = 0, MAX_PLAYERS - 1 do
    table.insert(requestData, { timer = 0, bestOf = 1, maxSuddenDeathTimer = 3600, sky = 0 })
end

local bestOf = 1
local maxSuddenDeathTimer = 3600 -- 2 minutes

gLevelValues.floatingStarDance = 1
gServerSettings.playerInteractions = PLAYER_INTERACTIONS_PVP

---@param m MarioState
function mario_update(m)
    local np = gNetworkPlayers[m.playerIndex]
    if np.currLevelNum ~= LEVEL_ARENA then
        if m.playerIndex == 0 then
            prevLevel, prevAct, prevArea = np.currLevelNum, np.currActNum, np.currAreaIndex
        end
        return
    end

    if hostIndex == -1 then return end

    local sMario = gPlayerSyncTable[m.playerIndex]
    local hostSMario = gPlayerSyncTable[hostIndex]

    if m.action == ACT_STAR_DANCE_EXIT and m.actionTimer > 50 then
        m.actionTimer = 50
        if m.playerIndex == 0 then
            betweenTimer = betweenTimer + 1
            if betweenTimer > 100 then
                hostIndex = -1
                warp_to_level(prevLevel, prevArea, prevAct)
                gMarioStates[0].health = 0x880
                sMario.compID = -1
            elseif m.playerIndex == 0 and m.controller.buttonPressed & START_BUTTON ~= 0 then
                djui_open_pause_menu()
            end
        end
    end

    if sMario.compID == nil or sMario.compID == -1 then
        -- spectator stuff goes here
        set_mario_action(m, ACT_DEBUG_FREE_MOVE_CUSTOM, 0)
        if m.playerIndex == 0 then
            local emptyDuel = true
            for i = 1, MAX_PLAYERS - 1 do
                if is_player_active(gMarioStates[i]) ~= 0 and sMario.compID and sMario.compID ~= -1 then
                    emptyDuel = false
                    break
                end
            end

            if emptyDuel then
                warp_to_level(prevLevel, prevArea, prevAct)
            end

            if hostSMario.matchState == 2 then
                drop_queued_background_music()
                fadeout_level_music(1)
            end
        end
        return
    end

    if hostSMario.startCountdown > 0 then
        betweenTimer = 0
        m.health = 0x880
        m.pos.x = 0

        if sMario.compID == 1 then
            m.pos.z = -2000
        else
            m.pos.z = 2000
        end

        if (m.playerIndex == 0 and not didStart) or m.action == ACT_DO_NOTHING then
            m.pos.y = 300
            if sMario.compID == 1 then
                m.faceAngle.y = 0
            else
                m.faceAngle.y = 0x8000
            end
            set_mario_action(m, ACT_IDLE, 0)
            if m.playerIndex == 0 then -- reset camera
                set_camera_mode(m.area.camera, CAMERA_MODE_ROM_HACK, 0)
                reset_camera(m.area.camera)
                gLakituState.pos.x = m.pos.x + sins(m.faceAngle.y + 0x8000) * 2000
                gLakituState.pos.z = m.pos.z + coss(m.faceAngle.y + 0x8000) * 2000
                gLakituState.posHSpeed = 0
                gLakituState.posVSpeed = 0
                vec3f_copy(m.area.camera.pos, gLakituState.pos)
                vec3f_copy(gLakituState.curPos, gLakituState.pos)
                vec3f_copy(gLakituState.goalPos, gLakituState.pos)

                didStart = true
            end
        end

        return
    elseif hostSMario.suddenDeath and m.health > 0xFF and m.action ~= ACT_DO_NOTHING then
        m.health = 0x100
    end

    if m.playerIndex == 0 then
        local oppIndex = network_local_index_from_global(sMario.opponent)
        local oppM = gMarioStates[oppIndex]
        if np.currAreaSyncValid and is_player_active(oppM) == 0 then
            if opponentDCTimer == 0 or hostSMario.matchState == 2 then
                opponentDCTimer = 60
                hostIndex = -1
                warp_to_level(prevLevel, prevArea, prevAct)
                m.health = 0x880
                sMario.compID = -1
                if hostSMario.matchState ~= 2 then
                    djui_popup_create("Your opponent disconnected!", 2)
                end
                return
            end
            opponentDCTimer = opponentDCTimer - 1
        else
            opponentDCTimer = 60
        end

        if hostIndex == 0 and hostSMario.matchState == 1 and m.action == ACT_DO_NOTHING and oppM.action == ACT_DO_NOTHING then
            network_send_to_include_self(hostIndex, true, {
                id = PACKET_NEXT_ROUND,
                restart = true,
            })
            didStart = false
            djui_popup_create_this_level("Tie!", 2)
            return
        elseif hostSMario.matchState == 2 then
            drop_queued_background_music()
            fadeout_level_music(1)
        end

        if m.action ~= ACT_DO_NOTHING then
            if hostSMario.matchState == 0 and m.health < 0x100 then -- interrupt death with explosion
                m.actionTimer = m.actionTimer + 1
                if m.actionTimer > 7 then
                    local o = spawn_sync_object(id_bhvExplosion,
                        E_MODEL_EXPLOSION,
                        m.pos.x, m.pos.y, m.pos.z,
                        nil)
                    o.oIntangibleTimer = -1
                    m.marioObj.header.gfx.node.flags = m.marioObj.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
                    set_mario_action(m, ACT_DO_NOTHING, 0)

                    hostSMario.matchState = 1
                end
            elseif hostSMario.matchState == 1 then
                if betweenTimer == 0 then
                    gPlayerSyncTable[0].points = gPlayerSyncTable[0].points + 1
                    if gPlayerSyncTable[0].points > hostSMario.bestOf // 2 then
                        hostSMario.matchState = 2
                        betweenTimer = 0
                    end
                elseif betweenTimer == 90 then
                    network_send_to_include_self(hostIndex, true, {
                        id = PACKET_NEXT_ROUND,
                    })
                    didStart = false
                end
                betweenTimer = betweenTimer + 1
            elseif hostSMario.matchState == 2 then
                m.health = 0x880
                if m.action ~= ACT_FALL_AFTER_STAR_GRAB and m.action ~= ACT_STAR_DANCE_EXIT and m.action ~= ACT_STAR_DANCE_NO_EXIT then
                    set_mario_action(m, ACT_FALL_AFTER_STAR_GRAB, 0)
                    betweenTimer = 0
                end
            end
        end
    end
end

hook_event(HOOK_MARIO_UPDATE, mario_update)

function on_death(m)
    local np = gNetworkPlayers[m.playerIndex]
    if np.currLevelNum ~= LEVEL_ARENA or hostIndex == -1 then return true end
    local hostSMario = gPlayerSyncTable[hostIndex]
    if hostSMario.matchState == 0 and time_stop == TIME_STOP_MAX then
        m.health = 0xFF
        m.actionTimer = 7
    elseif m.health > 0xFF then
        m.vel.y = 0
    end
    return false
end

hook_event(HOOK_ON_DEATH, on_death)

-- slowdown effect + handle timers
local didFirstJoinStuff = false
function update()
    if not didFirstJoinStuff then
        local sMario = gPlayerSyncTable[0]
        sMario.matchState = 0
        sMario.startCountdown = 90
        sMario.suddenDeathTimer = 0
        sMario.maxSuddenDeathTimer = 3600
        sMario.round = 1
        sMario.suddenDeath = false
        sMario.bestOf = 1
        sMario.compID = -1
        didFirstJoinStuff = true
    end

    for player, data in ipairs(requestData) do
        if data.timer > 0 then
            data.timer = data.timer - 1
        end
    end

    if hostIndex == -1 then return end
    local hostSMario = gPlayerSyncTable[hostIndex]
    if hostIndex == 0 then
        if hostSMario.startCountdown > 0 then
            hostSMario.startCountdown = hostSMario.startCountdown - 1
        end
        if hostSMario.suddenDeathTimer > 0 and hostSMario.matchState == 0 then
            hostSMario.suddenDeathTimer = hostSMario.suddenDeathTimer - 1
            if hostSMario.suddenDeathTimer == 0 then
                local survivor = -1
                local dead = -1
                local aliveHealth = 0
                for i = 0, MAX_PLAYERS - 1 do
                    if gPlayerSyncTable[i].compID and gPlayerSyncTable[i].compID ~= -1 then
                        if gMarioStates[i].health > aliveHealth then
                            dead = survivor
                            survivor = i
                            aliveHealth = gMarioStates[i].health
                        elseif gMarioStates[i].health == aliveHealth then
                            survivor = -1
                            break
                        else
                            dead = i
                        end
                    end
                end

                if survivor == -1 then
                    djui_popup_create_this_level("Sudden Death!", 2)
                    hostSMario.suddenDeath = true
                else
                    djui_popup_create_this_level("Time out!", 2)
                    if dead ~= -1 then
                        network_send_to_include_self(dead, true, {
                            id = PACKET_DEATH,
                        })
                    end
                end
            end
        end
    end
    local np = gNetworkPlayers[0]
    if np.currLevelNum ~= LEVEL_ARENA then return end

    if hostSMario.sky == 1 then
        set_override_skybox(BACKGROUND_FLAMING_SKY)
    elseif hostSMario.sky == 2 then
        set_override_skybox(BACKGROUND_PURPLE_SKY)
    else
        set_override_skybox(BACKGROUND_BELOW_CLOUDS)
    end

    local timeStopGood = false
    if hostSMario.matchState == 0 then
        for i = 0, MAX_PLAYERS - 1 do
            local m = gMarioStates[i]
            if is_player_active(m) ~= 0 and m.health < 0x100 and m.action ~= ACT_DO_NOTHING then
                if time_stop ~= 0 then
                    enable_time_stop_including_mario()
                    time_stop = time_stop - 1
                else
                    disable_time_stop()
                    time_stop = TIME_STOP_MAX
                end
                timeStopGood = true
                break
            end
        end
    end
    if not timeStopGood then
        time_stop = TIME_STOP_MAX
        disable_time_stop()
    end
end

hook_event(HOOK_UPDATE, update)

function on_level_init()
    didStart = false
    local np = gNetworkPlayers[0]
    if np.currLevelNum ~= LEVEL_ARENA or hostIndex == -1 then
        set_override_skybox(-1)
        gPlayerSyncTable[0].compID = -1
        return
    end
    local hostSMario = gPlayerSyncTable[hostIndex]
    if hostSMario.sky == 1 then
        set_background_music(0, SEQ_LEVEL_BOSS_KOOPA, 0)
    elseif hostSMario.sky == 2 then
        set_background_music(0, SEQ_LEVEL_BOSS_KOOPA_FINAL, 0)
    else
        set_background_music(0, SEQ_LEVEL_GRASS, 0)
    end
end

hook_event(HOOK_ON_LEVEL_INIT, on_level_init)

-- all hud stuff
local comeFromSideTimer = 30
function on_hud_render()
    local np = gNetworkPlayers[0]
    if np.currLevelNum ~= LEVEL_ARENA or hostIndex == -1 then return end

    local hudvalue = hud_get_value(HUD_DISPLAY_FLAGS)
    hud_set_value(HUD_DISPLAY_FLAGS, hudvalue & ~(HUD_DISPLAY_FLAGS_LIVES | HUD_DISPLAY_FLAGS_STAR_COUNT))

    local comps = {}
    for i = 0, MAX_PLAYERS - 1 do
        if is_player_active(gMarioStates[i]) ~= 0 and gPlayerSyncTable[i].compID and gPlayerSyncTable[i].compID ~= -1 then
            table.insert(comps, i)
            --table.insert(comps, i)
        end
    end

    local hostSMario = gPlayerSyncTable[hostIndex]

    djui_hud_set_resolution(RESOLUTION_DJUI)
    djui_hud_set_font(FONT_NORMAL)

    local screenWidth = djui_hud_get_screen_width()
    local screenHeight = djui_hud_get_screen_height()
    local x = 0
    local y = 0
    local width = 0
    local height = 0

    if hostSMario.matchState == 0 then
        comeFromSideTimer = 30
    end

    if hostSMario.startCountdown and hostSMario.startCountdown > 0 then
        if hostSMario.startCountdown > 90 then
            width = screenWidth * 0.4
            height = screenHeight * 0.1
            x = 0
            y = (screenHeight - height) / 2
            for i, index in ipairs(comps) do
                local playerColor = network_get_player_text_color_string(index)
                local name = playerColor .. gNetworkPlayers[index].name
                local r, g, b = convert_color(playerColor)
                r = math.max(r - 50, 0)
                g = math.max(g - 50, 0)
                b = math.max(b - 50, 0)

                djui_hud_set_color(r, g, b, 180)
                djui_hud_render_rect(x, y, width, height)
                if i == 1 then
                    x = x + 10
                else
                    x = x + width - djui_hud_measure_text(remove_color(name)) * 2 - 10
                end
                djui_hud_print_text_with_color_and_shadow(name, x, y + 2, 2, 255, 4)
                x = screenWidth - width
            end

            local text = "Best of " .. hostSMario.bestOf
            if hostSMario.bestOf == 1 then
                text = "First To Win"
            end
            width = djui_hud_measure_text(text)
            x = (screenWidth - width) / 2
            y = screenHeight - 50
            djui_hud_print_text_with_color_and_shadow(text, x, y, 1, 255, 2)

            djui_hud_set_font(FONT_HUD)
            text = "VS"
            width = djui_hud_measure_text(text) * 4
            x = (screenWidth - width) / 2
            y = (screenHeight - 65) / 2
            djui_hud_print_text_with_color_and_shadow(text, x, y, 4, 255, 4)
        else
            djui_hud_set_font(FONT_MENU)
            local text = tostring(hostSMario.startCountdown // 30 + 1)
            width = djui_hud_measure_text(text) * 8
            x = (screenWidth - width) / 2
            y = (screenHeight - 260) / 2
            djui_hud_print_text_with_color_and_shadow(text, x, y, 4, 255)

            if hostSMario.bestOf ~= 1 then
                djui_hud_set_font(FONT_NORMAL)
                text = "Round " .. hostSMario.round
                if hostSMario.bestOf and hostSMario.bestOf % 2 ~= 0 and hostSMario.round == hostSMario.bestOf then
                    text = "Final Round!"
                elseif hostSMario.bestOf and hostSMario.round > hostSMario.bestOf then
                    text = "Tiebreaker Round!"
                end
                width = djui_hud_measure_text(text)
                x = (screenWidth - width) / 2
                y = screenHeight - 50
                djui_hud_print_text_with_color_and_shadow(text, x, y, 1, 255, 2)
            end
        end
    elseif hostSMario.matchState == 0 then
        if hostSMario.suddenDeathTimer <= 900 and hostSMario.suddenDeathTimer ~= 0 then
            local text = tostring(math.ceil(hostSMario.suddenDeathTimer / 30))
            width = djui_hud_measure_text(text) * 2
            height = 30 * 2

            x = (screenWidth - width) - 10
            y = (screenHeight - height) - 2
            djui_hud_set_color(255, 255, 255, 255)
            djui_hud_print_text_with_color_and_shadow(text, x, y, 2, 255, 4)
        end
        if hostSMario.bestOf and hostSMario.bestOf > 1 then
            x = 20
            y = 20
            for i, index in ipairs(comps) do
                for a = 0, (hostSMario.bestOf // 2) do
                    if gPlayerSyncTable[index].points > a then
                        djui_hud_set_color(255, 255, 255, 255)
                    else
                        djui_hud_set_color(0, 0, 0, 180)
                    end
                    djui_hud_render_texture(gTextures.star, x, y, 4, 4)
                    if i == 1 then
                        x = x + 64
                    else
                        x = x - 64
                    end
                end
                x = screenWidth - 84
            end
        end
    elseif hostSMario.matchState == 1 then
        width = screenWidth * 0.4
        height = screenHeight * 0.2
        local dist = comeFromSideTimer * comeFromSideTimer * (width / 900)
        x = -dist
        y = (screenHeight - height) / 2

        for i, index in ipairs(comps) do
            local playerColor = network_get_player_text_color_string(index)
            local name = playerColor .. gNetworkPlayers[index].name
            local r, g, b = convert_color(playerColor)
            r = math.max(r - 50, 0)
            g = math.max(g - 50, 0)
            b = math.max(b - 50, 0)

            djui_hud_set_color(r, g, b, 180)
            djui_hud_render_rect(x, y, width, height)
            if i == 1 then
                x = x + 10
            else
                x = x + width - djui_hud_measure_text(remove_color(name)) * 2 - 10
            end
            djui_hud_print_text_with_color_and_shadow(name, x, y + 2, 2, 255, 4)
            y = y + 70
            if i ~= 1 then
                x = screenWidth - 74 + dist
            end
            for a = 0, (hostSMario.bestOf // 2) do
                if gPlayerSyncTable[index].points > a then
                    djui_hud_set_color(255, 255, 255, 255)
                else
                    djui_hud_set_color(0, 0, 0, 180)
                end
                djui_hud_render_texture(gTextures.star, x, y, 4, 4)

                if i == 1 then
                    x = x + 70
                else
                    x = x - 70
                end
            end
            x = screenWidth - width + dist
            y = y - 70
        end

        if comeFromSideTimer > 0 then
            comeFromSideTimer = comeFromSideTimer - 1
        end
    elseif hostSMario.matchState == 2 then
        local won = 0
        for i, index in ipairs(comps) do
            if gPlayerSyncTable[index].points > hostSMario.bestOf // 2 then
                won = index
                break
            end
        end
        local playerColor = network_get_player_text_color_string(won)
        local name = playerColor .. gNetworkPlayers[won].name
        text = name .. "\\#ffffff\\ wins!"

        twidth = djui_hud_measure_text(remove_color(text)) * 2
        width = twidth + 20
        height = 70
        x = (screenWidth - width) / 2
        y = (screenHeight - height) / 2

        local r, g, b = convert_color(playerColor)
        r = math.max(r - 50, 0)
        g = math.max(g - 50, 0)
        b = math.max(b - 50, 0)
        djui_hud_set_color(r, g, b, 180)
        djui_hud_render_rect(x, y, width, height)

        x = x + 10
        y = y + 5
        djui_hud_print_text_with_color_and_shadow(text, x, y, 2, 255, 4)
    end
end

hook_event(HOOK_ON_HUD_RENDER, on_hud_render)

hook_mario_action(ACT_DO_NOTHING, function(m)
    -- guess what happens!
    m.marioObj.header.gfx.node.flags = m.marioObj.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
    if m.playerIndex == 0 and m.controller.buttonPressed & START_BUTTON ~= 0 then
        djui_open_pause_menu()
    end
end, 0)

-- debug free move
function act_debug_free_move_custom(m)
    if not m then
        return true
    end
    m.health = 0x880
    local action = ACT_IDLE
    local speed = (m.controller.buttonDown & B_BUTTON ~= 0) and 1 or 4
    --set_mario_animation(m, MARIO_ANIM_A_POSE)
    m.marioObj.header.gfx.node.flags = m.marioObj.header.gfx.node.flags | GRAPH_RENDER_INVISIBLE
    if m.playerIndex == 0 and m.controller.buttonPressed & START_BUTTON ~= 0 then
        djui_open_pause_menu()
    end

    if m.controller.buttonDown & A_BUTTON ~= 0 then
        m.pos.y = m.pos.y + 16.0 * speed
    end
    if m.controller.buttonDown & Z_TRIG ~= 0 then
        m.pos.y = m.pos.y - 16.0 * speed
    end
    if m.intendedMag > 0 then
        m.pos.x = m.pos.x + 26.0 * speed * sins(m.intendedYaw)
        m.pos.z = m.pos.z + 26.0 * speed * coss(m.intendedYaw)
    end

    resolve_and_return_wall_collisions(m.pos, 0, 100)
    if m.floorHeight > m.pos.y then
        m.pos.y = m.floorHeight
    end

    m.faceAngle.y = m.intendedYaw
    vec3f_copy(m.marioObj.header.gfx.pos, m.pos)
    vec3s_set(m.marioObj.header.gfx.angle, 0, m.faceAngle.y, 0)
    if m.controller.buttonPressed == L_TRIG and m.marioObj.oTimer > 10 then
        if m.pos.y <= m.waterLevel - 100 then
            action = ACT_WATER_IDLE
        end
        set_mario_action(m, action, 0)
    end
    return false
end

hook_mario_action(ACT_DEBUG_FREE_MOVE_CUSTOM, act_debug_free_move_custom, 0)

function best_command(msg)
    local num = tonumber(msg)

    if num and math.floor(num) == num and num > 0 and num < 100 then
        bestOf = tonumber(msg)
        djui_chat_message_create("Now is a best of " .. num)
        return true
    end
end

function sky_command(msg)
    local num = tonumber(msg)

    if num and math.floor(num) == num and num >= 0 and num <= 2 then
        gPlayerSyncTable[0].sky = tonumber(msg)
        djui_chat_message_create("Set sky to " .. num)
        return true
    end
end

function time_command(msg)
    local num = tonumber(msg)

    if num and math.floor(num) == num and num > 0 and num <= 600 then
        maxSuddenDeathTimer = tonumber(msg) * 30
        djui_chat_message_create("Set max round length to " .. num .. " seconds")
        return true
    end
end

function force_duel_command(msg)
    if not (network_is_server() or network_is_moderator()) then
        djui_chat_message_create("Woah there, pardner. You ain't got permission for that.")
        return true
    end

    local args = split(msg, " ", 2)
    if not (args[1] or args[2]) then
        djui_chat_message_create("Need to specify two players!")
        return true
    end

    local playerID, np = (args[1] ~= "random" and get_specified_player(args[1]))
    local playerID2, np2 = (args[2] ~= "random" and get_specified_player(args[2]))
    if args[1] == "random" then
        local valid = {}
        for i = 1, MAX_PLAYERS - 1 do
            local np = gNetworkPlayers[i]
            if i ~= playerID2 and np.currLevelNum ~= LEVEL_ARENA then
                table.insert(valid, i)
            end
        end

        if #valid == 0 then
            djui_chat_message_create("There's not enough people for random, rip")
            return true
        end

        playerID = valid[math.random(1, #valid)]
        np = gNetworkPlayers[playerID]
    end
    if not np then return true end
    if np.currLevelNum == LEVEL_ARENA then
        local name = remove_color(np.name)
        djui_chat_message_create(name .. " is currently in a duel.")
        return true
    end

    if args[2] == "random" then
        local valid = {}
        for i = 1, MAX_PLAYERS - 1 do
            local np = gNetworkPlayers[i]
            if i ~= playerID and np.currLevelNum ~= LEVEL_ARENA then
                table.insert(valid, i)
            end
        end

        if #valid == 0 then
            djui_chat_message_create("There's not enough people for random, rip")
            return true
        end

        playerID2 = valid[math.random(1, #valid)]
        np2 = gNetworkPlayers[playerID]
    end
    if not np2 then return true end
    if np2.currLevelNum == LEVEL_ARENA then
        local name = remove_color(np2.name)
        djui_chat_message_create(name .. " is currently in a duel.")
        return true
    end

    if playerID == playerID2 then
        djui_chat_message_create("confrontingyourself.mp3")
        return true
    end

    local sMario = gPlayerSyncTable[0]
    network_send_to_include_self(playerID, true, {
        id = PACKET_GAME_START,
        host = np.globalIndex,
        opponent = np2.globalIndex,
        bestOf = sMario.bestOf,
        maxSuddenDeathTimer = sMario.maxSuddenDeathTimer,
        sky = sMario.sky,
        compID = 0,
    })

    network_send_to_include_self(playerID2, true, {
        id = PACKET_GAME_START,
        host = np.globalIndex,
        opponent = np.globalIndex,
        compID = 1,
    })

    djui_chat_message_create(string.format("Forcing %s and %s to duel", remove_color(np.name), remove_color(np2.name)))
    return true
end

function spectate_duel_command(msg)
    if (msg == nil or msg:lower() == "off" or msg == "") then
        if gNetworkPlayers[0].currLevelNum == LEVEL_ARENA then
            warp_to_level(prevLevel, prevArea, prevAct)
        end
        return true
    end

    local playerID, np = get_specified_player(msg)
    if not playerID then return true end
    if playerID == 0 then
        djui_chat_message_create("You can't spectate yourself!")
        return true
    elseif np.currLevelNum ~= LEVEL_ARENA then
        djui_chat_message_create("That player isn't currently in a duel.")
        return true
    elseif gNetworkPlayers[0].currLevelNum == LEVEL_ARENA then
        djui_chat_message_create("You're already in a duel!")
        return true
    end

    hostIndex = network_local_index_from_global(np.currActNum)
    warp_to_level(LEVEL_ARENA, 1, np.currActNum)
    return true
end

function duel_command(msg)
    local playerID, np
    if msg == "random" then
        local valid = {}
        for i = 1, MAX_PLAYERS - 1 do
            local np = gNetworkPlayers[i]
            if np.connected and np.currLevelNum ~= LEVEL_ARENA then
                table.insert(valid, i)
            end
        end

        if #valid == 0 then
            djui_chat_message_create("Oh god, EVERYONE'S dueling right now.")
            return true
        end

        playerID = valid[math.random(1, #valid)]
        np = gNetworkPlayers[playerID]
    else
        playerID, np = get_specified_player(msg)
    end

    if not np then return true end
    if playerID == 0 then
        djui_chat_message_create("confrontingyourself.mp3")
        return true
    elseif np.currLevelNum == LEVEL_ARENA then
        djui_chat_message_create("That player is currently in a duel.")
        return true
    elseif gNetworkPlayers[0].currLevelNum == LEVEL_ARENA then
        djui_chat_message_create("You're already in a duel!")
        return true
    end

    network_send_to(playerID, true, {
        id = PACKET_GAME_REQUEST,
        from = gNetworkPlayers[0].globalIndex,
        bestOf = bestOf,
        maxSuddenDeathTimer = maxSuddenDeathTimer,
    })
    djui_chat_message_create(string.format("Request sent to %s!", remove_color(np.name)))
    return true
end

function accept_command(msg)
    local playerID, np = get_specified_player(msg)
    if not np then return true end
    if np.localIndex == 0 then
        djui_chat_message_create("confrontingyourself.mp3")
        return true
    elseif np.currLevelNum == LEVEL_ARENA then
        djui_chat_message_create("Too late, that player is already in a duel.")
        return true
    elseif gNetworkPlayers[0].currLevelNum == LEVEL_ARENA then
        djui_chat_message_create("You're already in a duel!")
        return true
    elseif requestData[playerID].timer == 0 then
        djui_chat_message_create("Request expired, or never existed...")
        return true
    end

    local playerID2, np2 = 0, gNetworkPlayers[0]

    local data = requestData[playerID]
    data.timer = 0
    network_send_to_include_self(playerID, true, {
        id = PACKET_GAME_START,
        host = np.globalIndex,
        opponent = np2.globalIndex,
        bestOf = data.bestOf,
        maxSuddenDeathTimer = data.maxSuddenDeathTimer,
        compID = 0,
    })

    network_send_to_include_self(playerID2, true, {
        id = PACKET_GAME_START,
        host = np.globalIndex,
        opponent = np.globalIndex,
        compID = 1,
    })
end

function forfeit_command()
    if gNetworkPlayers[0].currLevelNum ~= LEVEL_ARENA then
        djui_chat_message_create("You aren't even IN a duel!")
        return true
    end
    gMarioStates[0].health = 0x880
    warp_to_level(prevLevel, prevArea, prevAct)
    return true
end

hook_chat_command("forceduel", "[NAME|ID,NAME|ID] - Force two players to battle!", force_duel_command)
hook_chat_command("duel", "[NAME|ID] - Request to duel another player!", duel_command)
hook_chat_command("accept", "[NAME|ID] - Accept a request from another player.", accept_command)
hook_chat_command("bestof", "[NUM] - Set total rounds for the duel", best_command)
hook_chat_command("sky", "[NUM] - Change sky; 0 is normal, 1 is BITFS, 2 is BITS", sky_command)
hook_chat_command("time", "[NUM] - Set max round length in seconds", time_command)
hook_chat_command("spectate", "[NAME|ID] - Spectate a player who is in a duel!", spectate_duel_command)
hook_chat_command("forfeit", "- Leave the current duel", forfeit_command)

function on_game_start(data)
    didStart = false
    hostIndex = network_local_index_from_global(data.host)

    warp_to_level(LEVEL_ARENA, 1, data.host)
    local sMario = gPlayerSyncTable[0]
    sMario.compID = data.compID or 0
    sMario.points = 0
    sMario.opponent = data.opponent
    if hostIndex == 0 then
        sMario.startCountdown = 180
        sMario.maxSuddenDeathTimer = data.maxSuddenDeathTimer
        sMario.suddenDeathTimer = sMario.maxSuddenDeathTimer + sMario.startCountdown
        sMario.matchState = 0
        sMario.round = 1
        sMario.suddenDeath = false
        sMario.bestOf = data.bestOf
        sMario.sky = data.sky or sMario.sky
    else
        gPlayerSyncTable[hostIndex].matchState = 0
    end
end

function on_packet_death(data)
    on_death(gMarioStates[0])
end

function on_game_request(data)
    local np = network_player_from_global_index(data.from)
    local playerColor = network_get_player_text_color_string(np.localIndex)
    djui_chat_message_create("Game request from " .. playerColor .. np.name .. "\\#dcdcdc\\!")
    local text = "Best of " .. data.bestOf
    if data.bestOf == 1 then
        text = "First To Win"
    end
    djui_chat_message_create(text)
    local minutes = data.maxSuddenDeathTimer // 1800
    local seconds = (data.maxSuddenDeathTimer // 30) % 60
    djui_chat_message_create(string.format("Max time: %d:%02d", minutes, seconds))
    djui_chat_message_create("Type /accept " .. data.from .. " to accept the duel! (Expires in 60s)")
    requestData[np.localIndex].timer = 1800
    requestData[np.localIndex].bestOf = data.bestOf
    requestData[np.localIndex].maxSuddenDeathTimer = data.maxSuddenDeathTimer
end

function on_packet_next_round(data)
    local sMario = gPlayerSyncTable[0]
    sMario.startCountdown = 90
    sMario.suddenDeathTimer = sMario.maxSuddenDeathTimer + sMario.startCountdown
    sMario.matchState = 0
    if not data.restart then
        sMario.round = sMario.round + 1
    end
    sMario.suddenDeath = false
end

function on_packet_popup(data)
    djui_popup_create(data.message, data.lines)
end

PACKET_GAME_START = 1
PACKET_DEATH = 2
PACKET_GAME_REQUEST = 3
PACKET_NEXT_ROUND = 4
PACKET_POPUP = 5
local sPacketTable = {
    [PACKET_GAME_REQUEST] = on_game_request,
    [PACKET_GAME_START] = on_game_start,
    [PACKET_DEATH] = on_packet_death,
    [PACKET_NEXT_ROUND] = on_packet_next_round,
    [PACKET_POPUP] = on_packet_popup,
}
function on_packet_receive(data)
    if sPacketTable[data.id] then
        return sPacketTable[data.id](data)
    end
end

hook_event(HOOK_ON_PACKET_RECEIVE, on_packet_receive)

-- Converts string into a table using a determiner (but stop splitting after a certain amount)
function split(s, delimiter, limit_)
    local limit = limit_ or 999
    local result = {}
    local finalmatch = ""
    local i = 0
    for match in (s):gmatch(string.format("[^%s]+", delimiter)) do
        --djui_chat_message_create(match)
        i = i + 1
        if i >= limit then
            finalmatch = finalmatch .. match .. delimiter
        else
            table.insert(result, match)
        end
    end
    if i >= limit then
        finalmatch = string.sub(finalmatch, 1, string.len(finalmatch) - string.len(delimiter))
        table.insert(result, finalmatch)
    end
    return result
end

-- from MarioHunt
function get_specified_player(msg)
    local playerID = tonumber(msg)
    if msg == "" then
        return 0, gNetworkPlayers[0]
    end

    local np = nil
    if not playerID then
        for i = 0, (MAX_PLAYERS - 1) do
            np = gNetworkPlayers[i]
            if remove_color(np.name) == remove_color(msg) then
                playerID = i
                break
            end
        end

        if not playerID then
            local subname = remove_color(msg):lower()
            for i = 0, (MAX_PLAYERS - 1) do -- try sub name
                np = gNetworkPlayers[i]
                local name = remove_color(np.name):lower()
                if name:find(subname) then
                    playerID = i
                    break
                end
            end

            if not playerID then
                djui_chat_message_create("No such player: " .. msg)
                return nil
            end
        end
    elseif playerID ~= math.floor(playerID) or playerID < 0 or playerID > (MAX_PLAYERS - 1) then
        djui_chat_message_create("Invalid id: " .. msg)
        return nil
    else
        np = network_player_from_global_index(playerID)
        playerID = (np and np.localIndex)
    end
    if not (np and np.connected) then
        djui_chat_message_create("No such player: " .. msg)
        return nil
    end

    return playerID, np
end

-- prints text on the screen... with color!
function djui_hud_print_text_with_color_and_shadow(text, x, y, scale, alpha, thickness)
    local space = 0
    local color = ""
    local render = ""
    local r, g, b, a = 255, 255, 255, alpha or 255
    text, color, render = remove_color(text, true)
    while render do
        if thickness and thickness > 0 and render:len() > 0 then
            djui_hud_set_color(0, 0, 0, alpha)
            djui_hud_print_text(render, x + space - thickness, y, scale)
            djui_hud_print_text(render, x + space, y - thickness, scale)
            djui_hud_print_text(render, x + space + thickness, y, scale)
            djui_hud_print_text(render, x + space, y + thickness, scale)
        end

        if r then djui_hud_set_color(r, g, b, alpha or a) end
        if render:len() > 0 then
            djui_hud_print_text(render, x + space, y, scale)
        end
        r, g, b, a = convert_color(color)
        space = space + djui_hud_measure_text(render) * scale
        text, color, render = remove_color(text, true)
    end

    if thickness and thickness > 0 and text:len() > 0 then
        djui_hud_set_color(0, 0, 0, alpha)
        djui_hud_print_text(text, x + space - thickness, y, scale)
        djui_hud_print_text(text, x + space, y - thickness, scale)
        djui_hud_print_text(text, x + space + thickness, y, scale)
        djui_hud_print_text(text, x + space, y + thickness, scale)
    end

    if r then djui_hud_set_color(r, g, b, alpha or a) end
    djui_hud_print_text(text, x + space, y, scale);
end

-- removes color string
function remove_color(text, get_color)
    local start = text:find("\\")
    local next = 1
    while (next) and (start) do
        start = text:find("\\")
        if start then
            next = text:find("\\", start + 1)
            if not next then
                next = text:len() + 1
            end

            if get_color then
                local color = text:sub(start, next)
                local render = text:sub(1, start - 1)
                text = text:sub(next + 1)
                return text, color, render
            else
                text = text:sub(1, start - 1) .. text:sub(next + 1)
            end
        end
    end
    return text
end

-- converts hex string to RGB values
function convert_color(text)
    if text:sub(2, 2) ~= "#" then
      return nil
    end
    text = text:sub(3, -2)
    local rstring, gstring, bstring = "", "", ""
    if text:len() ~= 3 and text:len() ~= 6 then return 255, 255, 255, 255 end
    if text:len() == 6 then
      rstring = text:sub(1, 2) or "ff"
      gstring = text:sub(3, 4) or "ff"
      bstring = text:sub(5, 6) or "ff"
    else
      rstring = text:sub(1, 1) .. text:sub(1, 1)
      gstring = text:sub(2, 2) .. text:sub(2, 2)
      bstring = text:sub(3, 3) .. text:sub(3, 3)
    end
    local r = tonumber("0x" .. rstring) or 255
    local g = tonumber("0x" .. gstring) or 255
    local b = tonumber("0x" .. bstring) or 255
    return r, g, b, 255 -- alpha is no longer writeable
  end

-- gets around error when sending packet to self
function network_send_to_include_self(index, reliable, data)
    if index ~= 0 then
        return network_send_to(index, reliable, data)
    end
    on_packet_receive(data)
end

-- creates a popup for all players in the same level
function djui_popup_create_this_level(message, lines)
    for i = 1, MAX_PLAYERS - 1 do
        if is_player_active(gMarioStates[i]) ~= 0 then
            network_send_to(i, true, {
                id = PACKET_POPUP,
                message = message,
                lines = lines,
            })
        end
    end
    djui_popup_create(message, lines)
end
